from deciphon_core.cffi import ffi, lib

__all__ = ["DeciphonError"]


class DeciphonError(RuntimeError):
    def __init__(self, errno: int):
        msg = ffi.string(lib.dcp_strerror(errno)).decode()
        super().__init__(f"deciphon_core error: {msg}")



from pydantic import BaseModel

from deciphon_core.cffi import ffi

__all__ = ["ScanParams"]


class ScanParams(BaseModel):
    num_threads: int = 1
    lrt_threshold: float = 0.0
    multi_hits: bool = True
    hmmer3_compat: bool = False

    @property
    def cparams(self):
        params = ffi.new(
            "struct dcp_scan_params*",
            {
                "num_threads": self.num_threads,
                "lrt_threshold": self.lrt_threshold,
                "multi_hits": self.multi_hits,
                "hmmer3_compat": self.hmmer3_compat,
            },
        )
        if params == ffi.NULL:
            raise MemoryError()
        return params[0]



from __future__ import annotations

from typing import Iterator

from deciphon_core.cffi import ffi, lib
from deciphon_core.cseq import CSeqIter
from deciphon_core.error import DeciphonError
from deciphon_core.scan_params import ScanParams
from deciphon_core.schema import DBFile, NewSnapFile
from deciphon_core.seq import Seq

__all__ = ["Scan"]


class Scan:
    def __init__(self):
        self._cscan = lib.dcp_scan_new()
        if self._cscan == ffi.NULL:
            raise MemoryError()

    def dial(self, port: int = 51371):
        if rc := lib.dcp_scan_dial(self._cscan, port):
            raise DeciphonError(rc)

    def setup(self, params: ScanParams):
        if rc := lib.dcp_scan_setup(self._cscan, params.cparams):
            raise DeciphonError(rc)

    def run(self, dbfile: DBFile, seqit: Iterator[Seq], snap: NewSnapFile):
        cscan = self._cscan
        db = bytes(dbfile.path)
        it = CSeqIter(seqit)
        basename = str(snap.basename).encode()
        if rc := lib.dcp_scan_run(cscan, db, it.c_callback, it.c_self, basename):
            raise DeciphonError(rc)

        snap.make_archive()

    def __del__(self):
        if getattr(self, "_cscan", ffi.NULL) != ffi.NULL:
            lib.dcp_scan_del(self._cscan)



from __future__ import annotations

from dataclasses import dataclass
from typing import Iterator, Optional

from deciphon_core.cffi import ffi, lib
from deciphon_core.error import DeciphonError
from deciphon_core.seq import Seq

__all__ = ["CSeq", "CSeqIter"]


@dataclass
class CSeq:
    id: int
    name: bytes
    data: bytes
    ctx: CSeqIter

    def __init__(self, seq: Seq, ctx: CSeqIter):
        self.id = seq.id
        self.name = seq.name.encode()
        self.data = seq.data.encode()
        self.ctx = ctx


class CSeqIter:
    def __init__(self, iter: Iterator[Seq]):
        self._cself = ffi.new_handle(self)
        self._iter = iter
        self._curr_seq: Optional[CSeq] = None

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self._curr_seq = None

    def __next__(self) -> CSeq:
        # Keep the current returned sequence alive during
        # the lifetime of `self`.
        self._curr_seq = CSeq(next(self._iter), self)
        return self._curr_seq

    def __iter__(self):
        return self

    @property
    def c_callback(self):
        return lib.next_seq_callb

    @property
    def c_self(self):
        return self._cself


@ffi.def_extern()
def next_seq_callb(cseq, cself):
    iter: CSeqIter = ffi.from_handle(cself)
    try:
        seq: CSeq = next(iter)
        if rc := lib.dcp_seq_setup(cseq, seq.id, seq.name, seq.data):
            raise DeciphonError(rc)
    except Exception:
        return False
    return True
